"""Get data into JVM for prediction and out again as Spark Dataframe"""
import logging
logger = logging.getLogger('nlu')
import pyspark
from pyspark.sql.functions import monotonically_increasing_id
import numpy as np
import pandas as pd
from pyspark.sql.types import StringType, StructType,StructField

class DataConversionUtils():
    # Modin aswell but optional, so we dont import the type yet
    supported_types = [pyspark.sql.DataFrame,pd.DataFrame, pd.Series,np.ndarray ]
    @staticmethod
    def except_text_col_not_found(cols):
        print(f'Could not find column named "text" in input Pandas Dataframe. Please ensure one column named such exists. Columns in DF are : {cols} ')
    @staticmethod
    def sdf_to_sdf(data,spark_sess,raw_text_column='text'):
        """No casting, Spark to Spark. Just add index col"""
        output_datatype = 'spark'
        data = data.withColumn('origin_index',monotonically_increasing_id().alias('origin_index'))
        stranger_features= []
        if raw_text_column in data.columns:
            # store all stranger features
            if len(data.columns) > 1:
                stranger_features = list(set(data.columns) - set(raw_text_column))
            else: DataConversionUtils.except_text_col_not_found(data.columns)
        return data, stranger_features, output_datatype

    @staticmethod
    def pdf_to_sdf(data,spark_sess,raw_text_column='text'):
        """Casting pandas to spark and add index col"""
        output_datatype = 'pandas'
        stranger_features= []
        sdf = None
        # set first col as text column if there is none
        if raw_text_column not in data.columns:data.rename(columns={data.columns[0]: 'text'}, inplace=True)
        data['origin_index'] = data.index
        if raw_text_column in data.columns:
            if len(data.columns) > 1:
                # make  Nans to None, or spark will crash
                data = data.where(pd.notnull(data), None)
                data = data.dropna(axis=1, how='all')
                stranger_features = list(set(data.columns) - set(raw_text_column))
            sdf = spark_sess.createDataFrame(data)
        else: DataConversionUtils.except_text_col_not_found(data.columns)
        return sdf, stranger_features, output_datatype

    @staticmethod
    def pds_to_sdf(data,spark_sess,raw_text_column='text'):
        """Casting pandas series to spark and add index col.  # for df['text'] colum/series passing casting follows pseries->pdf->spark->pd """
        output_datatype = 'pandas_series'
        sdf=None
        schema = StructType([StructField(raw_text_column,StringType(),True)])
        data = pd.DataFrame(data).dropna(axis=1, how='all')
        # If series from a column is passed, its column name will be reused.
        if raw_text_column not in data.columns and len(data.columns) == 1:
            data[raw_text_column] = data[data.columns[0]]
        else:
            logger.info(f'INFO: NLU will assume {data.columns[0]} as label column since default text column could not be find')
            data[raw_text_column] = data[data.columns[0]]
        data['origin_index'] = data.index
        if raw_text_column in data.columns:
            sdf = spark_sess.createDataFrame(pd.DataFrame(data[raw_text_column]),schema=schema)
        else: DataConversionUtils.except_text_col_not_found(data.columns)
        return sdf, [], output_datatype

    @staticmethod
    def np_to_sdf(data,spark_sess,raw_text_column='text'):
        """Casting numpy array to spark and add index col. This is a bit inefficient. Casting follow  np->pd->spark->pd. We could cut out the first pd step   """
        output_datatype = 'numpy_array'
        if len(data.shape) != 1: ValueError(f"Exception : Input numpy array must be 1 Dimensional for prediction.. Input data shape is{data.shape}")
        sdf = spark_sess.createDataFrame(pd.DataFrame({raw_text_column: data, 'origin_index': list(range(len(data)))}))
        return sdf, [], output_datatype

    @staticmethod
    def str_to_sdf(data,spark_sess,raw_text_column='text'):
        """Casting str  to spark and add index col. This is a bit inefficient. Casting follow  # inefficient, str->pd->spark->pd , we can could first pd"""
        output_datatype = 'string'
        sdf = spark_sess.createDataFrame(pd.DataFrame({raw_text_column: data, 'origin_index': [0]}, index=[0]))
        return sdf, [], output_datatype

    @staticmethod
    def str_list_to_sdf(data,spark_sess,raw_text_column='text'):
        """Casting str list  to spark and add index col. This is a bit inefficient. Casting follow  # # inefficient, list->pd->spark->pd , we can could first pd"""
        output_datatype = 'string_list'
        if all(type(elem) == str for elem in data):
            sdf = spark_sess.createDataFrame(pd.DataFrame({raw_text_column: pd.Series(data), 'origin_index': list(range(len(data)))}))
        else:
            ValueError("Exception: Not all elements in input list are of type string.")
        return sdf, [], output_datatype


    @staticmethod
    def fallback_modin_to_sdf(data,spark_sess,raw_text_column='text'):
        """Casting potential Modin data to spark and add index col. # Modin tests, This could crash if Modin not installed """
        sdf = None
        output_datatype = ''
        try:
            import modin.pandas as mpd
            if isinstance(data, mpd.DataFrame):
                data = pd.DataFrame(data.to_dict())  # create pandas to support type inference
                output_datatype = 'modin'
                data['origin_index'] = data.index
            if raw_text_column in data.columns:
                if len(data.columns) > 1:
                    data = data.where(pd.notnull(data), None)  # make  Nans to None, or spark will crash
                    data = data.dropna(axis=1, how='all')
                    stranger_features = list(set(data.columns) - set(raw_text_column))
                sdf = spark_sess.createDataFrame(data)
            else: DataConversionUtils.except_text_col_not_found(data.columns)
            if isinstance(data, mpd.Series):
                output_datatype = 'modin_series'
                data = pd.Series(data.to_dict())  # create pandas to support type inference
                data = pd.DataFrame(data).dropna(axis=1, how='all')
                data['origin_index'] = data.index
                index_provided = True
                if raw_text_column in data.columns:
                    sdf = spark_sess.createDataFrame(data[['text']])
                else: DataConversionUtils.except_text_col_not_found(data.columns)
        except: print("If you use Modin, make sure you have installed 'pip install modin[ray]' or 'pip install modin[dask]' backend for Modin ")
        return sdf, [], output_datatype


    @staticmethod
    def to_spark_df(data,spark_sess,raw_text_column='text') :
        """Convert supported datatypes to SparkDF and extract extra data for prediction later on."""
        try :
            if   isinstance(data,pyspark.sql.dataframe.DataFrame): return DataConversionUtils.sdf_to_sdf(data,spark_sess,raw_text_column)
            elif isinstance(data,pd.DataFrame):  return DataConversionUtils.pdf_to_sdf(data,spark_sess,raw_text_column)
            elif isinstance(data,pd.Series):     return DataConversionUtils.pds_to_sdf(data,spark_sess,raw_text_column)
            elif isinstance(data,np.ndarray):  return DataConversionUtils.np_to_sdf(data,spark_sess,raw_text_column)
            elif isinstance(data,str): return DataConversionUtils.str_to_sdf(data,spark_sess,raw_text_column)
            elif isinstance(data,list): return DataConversionUtils.str_list_to_sdf(data,spark_sess,raw_text_column)
            else: return DataConversionUtils.fallback_modin_to_sdf(data,spark_sess,raw_text_column)
        except : ValueError("Data could not be converted to Spark Dataframe for internal conversion.")


import nlu
from nlu.discovery import Discoverer
from nlu.pipe.utils.storage_ref_utils import StorageRefUtils
from typing import List, Tuple, Optional, Dict, Union
import streamlit as st
from nlu.utils.modelhub.modelhub_utils import ModelHubUtils
import numpy as np
import pandas as pd
from nlu.pipe.viz.streamlit_viz.streamlit_utils_OS import StreamlitUtilsOS
from nlu.pipe.viz.streamlit_viz.gen_streamlit_code import get_code_for_viz
from nlu.pipe.viz.streamlit_viz.styles import _set_block_container_style
import random
from nlu.pipe.viz.streamlit_viz.streamlit_viz_tracker import StreamlitVizTracker
class WordEmbeddingManifoldStreamlitBlock():
    @staticmethod
    def viz_streamlit_word_embed_manifold(
            pipe, # nlu pipe
            default_texts: List[str] = ("Donald Trump likes to party!", "Angela Merkel likes to party!", 'Peter HATES TO PARTTY!!!! :('),
            title: Optional[str] = "Lower dimensional Manifold visualization for word embeddings",
            sub_title: Optional[str] = "Apply any of the 11 `Manifold` or `Matrix Decomposition` algorithms to reduce the dimensionality of `Word Embeddings` to `1-D`, `2-D` and `3-D` ",
            write_raw_pandas : bool = False ,
            default_algos_to_apply : List[str] = ("TSNE", "PCA"),#,'LLE','Spectral Embedding','MDS','ISOMAP','SVD aka LSA','DictionaryLearning','FactorAnalysis','FastICA','KernelPCA',),  # LatentDirichletAllocation 'NMF',
            target_dimensions : List[int] = (1,2,3),
            show_algo_select : bool = True,
            show_embed_select : bool = True,
            show_color_select: bool = True,
            MAX_DISPLAY_NUM:int=100,
            display_embed_information:bool=True,
            set_wide_layout_CSS:bool=True,
            num_cols: int = 3,
            model_select_position:str = 'side', # side or main
            key:str = "NLU_streamlit",
            additional_classifiers_for_coloring:List[str]=['pos', 'sentiment.imdb'],
            generate_code_sample:bool = False,
            show_infos:bool = True,
            show_logo:bool = True,
            n_jobs: Optional[int] = 3, # False
    ):
        from nlu.pipe.viz.streamlit_viz.streamlit_utils_OS import StreamlitUtilsOS
        StreamlitVizTracker.footer_displayed=False
        try :
            import plotly.express as px
            from sklearn.metrics.pairwise import distance_metrics
        except :st.error("You need the sklearn and plotly package in your Python environment installed for similarity visualizations. Run <pip install sklearn plotly>")
        if len(default_texts) > MAX_DISPLAY_NUM : default_texts = default_texts[:MAX_DISPLAY_NUM]
        if show_logo :StreamlitVizTracker.show_logo()
        if set_wide_layout_CSS : _set_block_container_style()
        if title:st.header(title)
        if sub_title:st.subheader(sub_title)
        # if show_logo :VizUtilsStreamlitOS.show_logo()

        # VizUtilsStreamlitOS.loaded_word_embeding_pipes = []


        data = st.text_area('Enter N texts, seperated by new lines to visualize Word Embeddings for ','\n'.join(default_texts))
        if len(data) > MAX_DISPLAY_NUM : data = data[:MAX_DISPLAY_NUM]
        data_split = data.split("\n")
        while '' in data_split : data_split.remove('')
        data       = data_split.copy()
        while '' in data : data.remove('')

        if len(data)<=1:
            st.error("Please enter more than 2 lines of text, seperated by new lines (hit <ENTER>)")
            return
        # TODO dynamic color inference for plotting
        if show_color_select:
            if model_select_position == 'side' : feature_to_color_by =  st.sidebar.selectbox('Pick a feature to color points in manifold by ',['pos','sentiment',],0)
            else:feature_to_color_by =  st.selectbox('Feature to color plots by ',['pos','sentiment',],0)
        text_col = 'token'
        embed_algos_to_load = []
        new_embed_pipes = []
        e_coms = StreamlitUtilsOS.find_all_embed_components(pipe)

        if show_algo_select :
            exp = st.beta_expander("Select additional manifold and dimension reduction techniques to apply")

            algos = exp.multiselect(
                "Reduce embedding dimensionality to something visualizable",
                options=("TSNE", "ISOMAP",'LLE','Spectral Embedding','MDS','PCA','SVD aka LSA','DictionaryLearning','FactorAnalysis','FastICA','KernelPCA',),default=default_algos_to_apply,)

            emb_components_usable = [e for e in Discoverer.get_components('embed',True, include_aliases=True) if 'chunk' not in e and 'sentence' not in e]
            loaded_embed_nlu_refs = []
            loaded_classifier_nlu_refs = []
            loaded_storage_refs = []
            for c in e_coms :
                if not  hasattr(c.info,'nlu_ref'): continue
                r = c.info.nlu_ref
                if 'en.' not in r and 'embed.' not  in r and 'ner' not in r : loaded_embed_nlu_refs.append('en.embed.' + r)
                elif 'en.'  in r and 'embed.' not  in r  and 'ner' not in r:
                    r = r.split('en.')[0]
                    loaded_embed_nlu_refs.append('en.embed.' + r)
                else :
                    loaded_embed_nlu_refs.append(StorageRefUtils.extract_storage_ref(c))
                loaded_storage_refs.append(StorageRefUtils.extract_storage_ref(c))

            for p in StreamlitVizTracker.loaded_word_embeding_pipes :
                if p != pipe : loaded_embed_nlu_refs.append(p.nlu_ref)
            loaded_embed_nlu_refs = list(set(loaded_embed_nlu_refs))

            for l in loaded_embed_nlu_refs:
                if l not in emb_components_usable : emb_components_usable.append(l)
            emb_components_usable.sort()
            loaded_embed_nlu_refs.sort()
            if model_select_position =='side':
                embed_algo_selection   = st.sidebar.multiselect("Pick additional Word Embeddings for the Dimension Reduction",options=emb_components_usable,default=loaded_embed_nlu_refs,key = key)
                embed_algo_selection=[embed_algo_selection[-1]]
            else :
                exp = st.beta_expander("Pick additional Word Embeddings")
                embed_algo_selection   = exp.multiselect("Pick additional Word Embeddings for the Dimension Reduction",options=emb_components_usable,default=loaded_embed_nlu_refs,key = key)
                embed_algo_selection=[embed_algo_selection[-1]]
            embed_algos_to_load = list(set(embed_algo_selection) - set(loaded_embed_nlu_refs))
        for embedder in embed_algos_to_load:new_embed_pipes.append(nlu.load(embedder))# + f' {" ".join(additional_classifiers_for_coloring)}'))
        StreamlitVizTracker.loaded_word_embeding_pipes+=new_embed_pipes
        if pipe not in StreamlitVizTracker.loaded_word_embeding_pipes: StreamlitVizTracker.loaded_word_embeding_pipes.append(pipe)
        for nlu_ref in additional_classifiers_for_coloring :
            already_loaded=False
            if 'pos' in nlu_ref : continue
            # for p in  VizUtilsStreamlitOS.loaded_document_classifier_pipes:
            #     if p.nlu_ref == nlu_ref : already_loaded = True
            # if not already_loaded : VizUtilsStreamlitOS.loaded_token_level_classifiers.append(nlu.load(nlu_ref))
            else :
                for p in  StreamlitVizTracker.loaded_document_classifier_pipes:
                    if p.nlu_ref == nlu_ref : already_loaded = True
                if not already_loaded :
                    already_loaded=True
                    StreamlitVizTracker.loaded_document_classifier_pipes.append(nlu.load(nlu_ref))

        col_index = 0
        cols = st.beta_columns(num_cols)
        def are_cols_full(): return col_index == num_cols
        token_feature_pipe = StreamlitUtilsOS.get_pipe('pos')
        #not all pipes have sentiment/pos etc.. models for hueing loaded....
        ## Lets FIRST predict with the classifiers/Token level feature generators and THEN apply embed pipe
        for p in StreamlitVizTracker.loaded_word_embeding_pipes :
            data = data_split.copy()

            classifier_cols = []

            for class_p in StreamlitVizTracker.loaded_document_classifier_pipes:

                data = class_p.predict(data, output_level='document',multithread=False).dropna()
                classifier_cols.append(StreamlitUtilsOS.get_classifier_cols(class_p))
                data['text'] = data_split
                # drop embeds of classifiers because bad conversion
                for c in data.columns :
                    if 'embedding' in c : data.drop(c, inplace=True,axis=1)
            # data['text']
            # =data['document']
            data['text'] = data_split
            for c in data.columns :
                if 'sentence_embedding' in c : data.drop(c,inplace=True,axis=1)

            p = StreamlitUtilsOS.merge_token_classifiers_with_embed_pipe(p, token_feature_pipe)
            # if'token' in data.columns : data.drop(['token','pos'],inplace=True,)
            if'pos' in data.columns : data.drop('pos',inplace=True,axis=1)
            predictions =   p.predict(data,output_level='token',multithread=False).dropna()
            e_col = StreamlitUtilsOS.find_embed_col(predictions)
            e_com = StreamlitUtilsOS.find_embed_component(p)
            e_com_storage_ref = StorageRefUtils.extract_storage_ref(e_com, True)
            # embedder_name = StreamlitUtilsOS.extract_name(e_com)
            emb = predictions[e_col]
            mat = np.array([x for x in emb])
            for algo in algos :
                if len(mat.shape)>2 : mat = mat.reshape(len(emb),mat.shape[-1])
                hover_data = ['token','text','sentiment', 'pos']  # TODO DEDUCT
                # calc reduced dimensionality with every algo
                if 1 in target_dimensions:
                    low_dim_data = StreamlitUtilsOS.get_manifold_algo(algo,1,n_jobs).fit_transform(mat)
                    x = low_dim_data[:,0]
                    y = np.zeros(low_dim_data[:,0].shape)

                    tsne_df =  pd.DataFrame({'x':x,'y':y, 'text':predictions[text_col], 'pos':predictions.pos, 'sentiment' : predictions.sentiment,'token':predictions.token})
                    fig = px.scatter(tsne_df, x="x", y="y",color=feature_to_color_by, hover_data=hover_data)
                    subh = f"""Word-Embeddings =`{e_com_storage_ref}`, Manifold-Algo =`{algo}` for `D=1`"""
                    cols[col_index].markdown(subh)
                    cols[col_index].write(fig,key=key)
                    col_index+=1
                    if are_cols_full() :
                        cols = st.beta_columns(num_cols)
                        col_index = 0
                if 2 in target_dimensions:
                    low_dim_data = StreamlitUtilsOS.get_manifold_algo(algo,2,n_jobs).fit_transform(mat)
                    x = low_dim_data[:,0]
                    y = low_dim_data[:,1]
                    tsne_df =  pd.DataFrame({'x':x,'y':y, 'text':predictions[text_col], 'pos':predictions.pos, 'sentiment' : predictions.sentiment,'token':predictions.token})

                    fig = px.scatter(tsne_df, x="x", y="y",color=feature_to_color_by, hover_data=hover_data)
                    subh = f"""Word-Embeddings =`{e_com_storage_ref}`, Manifold-Algo =`{algo}` for `D=2`"""
                    cols[col_index].markdown(subh)
                    cols[col_index].write(fig,key=key)
                    # st.write(fig)
                    col_index+=1
                    if are_cols_full() :
                        cols = st.beta_columns(num_cols)
                        col_index = 0
                if 3 in target_dimensions:
                    low_dim_data = StreamlitUtilsOS.get_manifold_algo(algo,3,n_jobs).fit_transform(mat)
                    x = low_dim_data[:,0]
                    y = low_dim_data[:,1]
                    z = low_dim_data[:,2]
                    tsne_df =  pd.DataFrame({'x':x,'y':y,'z':z, 'text':predictions[text_col], 'pos':predictions.pos, 'sentiment':predictions.sentiment,'token':predictions.token })

                    fig = px.scatter_3d(tsne_df, x="x", y="y", z='z',color=feature_to_color_by, hover_data=hover_data)
                    subh = f"""Word-Embeddings =`{e_com_storage_ref}`, Manifold-Algo =`{algo}` for `D=3`"""
                    cols[col_index].markdown(subh)
                    cols[col_index].write(fig,key=key)

                    # st.write(fig)
                    col_index+=1
                    if are_cols_full() :
                        cols = st.beta_columns(num_cols)
                        col_index = 0
            # Todo fancy embed infos etc
            # if display_embed_information: display_embed_vetor_information(e_com,mat)

        # if display_embed_information:
        #     exp = st.beta_expander("Embedding vector information")
        #     exp.write(embed_vector_info)
        if show_infos :
            # VizUtilsStreamlitOS.display_infos()
            StreamlitVizTracker.display_model_info(pipe.nlu_ref, pipes = [pipe])
            StreamlitVizTracker.display_footer()


from sparknlp.annotator import *

class Marian:

    @staticmethod
    def get_default_model():
        return MarianTransformer.pretrained() \
            .setInputCols("document") \
            .setOutputCol("marian")

    @staticmethod
    def get_pretrained_model(name, language):
        return MarianTransformer.pretrained(name, language) \
            .setInputCols("document") \
            .setOutputCol("marian")






# from nlu import *
import nlu
class NLUComponent():
    '''
        This class loads all the components in the components folder.
        It acts as an accessor to every component.
        Pass a string identifier to it and it will construct and return the model.
        It will also take care of setting the initial configuration parameter dict .
     '''

    def __init__(self, component_name, component_type):
        # because we ae using the input/output/label fix, we run supers init AFTER model was created, so the next statement would overwrite it
        # when NLU might support 3rd party models we should rework this
        # self.model = None  # Either Spark NLP model or some 3rd party custom model. Reference to a model
        self.component_path = nlu.nlu_package_location + 'components/' + component_type + 's/' + component_name + '/'
        self.info = nlu.ComponentInfo.from_directory(component_info_dir=self.component_path)

    def print_parameters_explanation(self):
        pass

    def print_parameters(self):
        pass

    def info(self):
        print(self.info['info'])
        self.print_parameters_explanation()
        self.print_parameters()


class  SparkNLUComponent(NLUComponent):
    def __init__(self, component_name, component_type, nlu_ref='', nlp_ref='',lang='',loaded_from_pretrained_pipe=False, is_licensed=False):
        NLUComponent.__init__(self, component_name, component_type)
        self.info.nlu_ref = nlu_ref
        self.info.nlp_ref = nlp_ref
        self.info.lang    = lang
        self.info.loaded_from_pretrained_pipe = loaded_from_pretrained_pipe
        self.__set_missing_model_attributes__()
        if is_licensed : self.info.license = 'healthcare'

    def __set_missing_model_attributes__(self):
        '''
        For a given Spark NLP model this model will extract the poarameter map and search for input/output/label columns and set them on the model.
        This is a workaround to undefined behaviour when getting input/output/label columns
        :param : The model for which the attributes should be set
        :return: model with attributes properly set
        '''
        for k in self.model.extractParamMap():
            if "inputCol" in str(k):
                if isinstance(self.model.extractParamMap()[k], str) :
                    if self.model.extractParamMap()[k] == 'embeddings': # swap name so we have uniform col names
                        self.model.setInputCols( 'word_embeddings')
                    self.info.spark_input_column_names =  [self.model.extractParamMap()[k]]
                else :
                    if 'embeddings' in self.model.extractParamMap()[k]: # swap name so we have uniform col names
                        new_cols = self.model.extractParamMap()[k]
                        new_cols.remove("embeddings")
                        new_cols.append("word_embeddings")
                        self.model.setInputCols(new_cols)
                    self.info.spark_input_column_names =  self.model.extractParamMap()[k]

            if "outputCol" in str(k):
                if isinstance(self.model.extractParamMap()[k], str) :
                    if self.model.extractParamMap()[k] == 'embeddings': # swap name so we have uniform col names
                        self.model.setOutputCol( 'word_embeddings')
                    self.info.spark_output_column_names =  [self.model.extractParamMap()[k]]
                else :
                    if 'embeddings' in self.model.extractParamMap()[k]: # swap name so we have uniform col names
                        new_cols = self.model.extractParamMap()[k]
                        new_cols.remove("embeddings")
                        new_cols.append("word_embeddings")
                        self.model.setOutputCol(new_cols)
                    self.info.spark_output_column_names =  self.model.extractParamMap()[k]





            # if "labelCol" in str(k):
            #     if isinstance(self.model.extractParamMap()[k], str) :
            #         self.component_info['spark_label_column_names'] =  [self.model.extractParamMap()[k]]
            #     else :
            #         self.component_info['spark_label_column_names'] =  self.model.extractParamMap()[k]



import nlu
from nlu.discovery import Discoverer
from nlu.pipe.utils.storage_ref_utils import StorageRefUtils
from typing import List, Tuple, Optional, Dict, Union
import streamlit as st
from nlu.utils.modelhub.modelhub_utils import ModelHubUtils
import numpy as np
import pandas as pd
from nlu.pipe.viz.streamlit_viz.streamlit_utils_OS import StreamlitUtilsOS
from nlu.pipe.viz.streamlit_viz.gen_streamlit_code import get_code_for_viz
from nlu.pipe.viz.streamlit_viz.styles import _set_block_container_style
import random
from nlu.pipe.viz.streamlit_viz.streamlit_viz_tracker import StreamlitVizTracker
class ClassifierStreamlitBlock():

    @staticmethod
    def visualize_classes(
            pipe, # nlu pipe
            text:Union[str,list,pd.DataFrame, pd.Series, List[str]]=('I love NLU and Streamlit and sunny days!', 'I hate rainy daiys','CALL NOW AND WIN 1000$M'),
            output_level:Optional[str]='document',
            title: Optional[str] = "Text Classification",
            sub_title: Optional[str] = 'View predicted `classes` and `confidences` for `hundreds of text classifiers` in `over 200 languages`',
            metadata : bool = False,
            positions : bool = False,
            set_wide_layout_CSS:bool=True,
            generate_code_sample:bool = False,
            key:str = "NLU_streamlit",
            show_model_selector : bool = True ,
            model_select_position:str = 'side' ,
            show_infos:bool = True,
            show_logo:bool = True,
    )->None:
        if show_logo :StreamlitVizTracker.show_logo()
        if set_wide_layout_CSS : _set_block_container_style()
        if title:st.header(title)
        if sub_title:st.subheader(sub_title)

        # if generate_code_sample: st.code(get_code_for_viz('CLASSES',StreamlitUtilsOS.extract_name(pipe),text))
        if not isinstance(text, (pd.DataFrame, pd.Series)):
            text = st.text_area('Enter N texts, seperated by new lines to view classification results for','\n'.join(text) if isinstance(text,list) else text, key=key)
            text = text.split("\n")
            while '' in text : text.remove('')
        classifier_pipes = [pipe]
        classifier_components_usable = [e for e in Discoverer.get_components('classify',True, include_aliases=True)]
        classifier_components = StreamlitUtilsOS.find_all_classifier_components(pipe)
        loaded_classifier_nlu_refs = [c.info.nlu_ref for c in classifier_components ]

        for l in loaded_classifier_nlu_refs:
            if 'converter' in l :
                loaded_classifier_nlu_refs.remove(l)
                continue
            if l not in classifier_components_usable : classifier_components_usable.append(l)

        classifier_components_usable.sort()
        loaded_classifier_nlu_refs.sort()
        for r in loaded_classifier_nlu_refs:
            if r not in  classifier_components_usable : loaded_classifier_nlu_refs.remove(r)
        if show_model_selector :
            if model_select_position =='side':classifier_components_selection   = st.sidebar.multiselect("Pick additional Classifiers",options=classifier_components_usable,default=loaded_classifier_nlu_refs,key = key)
            else:classifier_components_selection   = st.multiselect("Pick additional Classifiers",options=classifier_components_usable,default=loaded_classifier_nlu_refs,key = key)
        # else : ValueError("Please define model_select_position as main or side")
        classifier_algos_to_load = list(set(classifier_components_selection) - set(loaded_classifier_nlu_refs))
        for classifier in classifier_algos_to_load:classifier_pipes.append(nlu.load(classifier))
        StreamlitVizTracker.loaded_document_classifier_pipes+= classifier_pipes
        if generate_code_sample:st.code(get_code_for_viz('CLASSES',[StreamlitUtilsOS.extract_name(p) for p  in classifier_pipes],text))

        dfs = []
        all_classifier_cols=[]
        for p in classifier_pipes :
            df = p.predict(text, output_level=output_level, metadata=metadata, positions=positions)
            classifier_cols = StreamlitUtilsOS.get_classifier_cols(p)
            for c in classifier_cols :
                if c not in df.columns : classifier_cols.remove(c)

            if 'text' in df.columns: classifier_cols += ['text']
            elif 'document' in df.columns: classifier_cols += ['document']
            all_classifier_cols+= classifier_cols
            dfs.append(df)
        df = pd.concat(dfs, axis=1)
        df = df.loc[:,~df.columns.duplicated()]
        for c in all_classifier_cols :
            if c not in df.columns : all_classifier_cols.remove(c)
        all_classifier_cols = list(set(all_classifier_cols))

        if len(all_classifier_cols) == 0: st.error('No classes detected')
        else :st.write(df[all_classifier_cols],key=key)
        if show_infos :
            # VizUtilsStreamlitOS.display_infos()
            StreamlitVizTracker.display_model_info(pipe.nlu_ref, pipes = [pipe])
            StreamlitVizTracker.display_footer()








import unittest
from nlu import *

class PretrainedPipeTests(unittest.TestCase):

    def simple_pretrained_pipe_tests(self):
        df = nlu.load('ner.onto',verbose=True).predict('I love peanutbutter and jelly')
        for c in df.columns: print(df[c])

if __name__ == '__main__':
    unittest.main()




from sparknlp_jsl.annotator import ChunkMergeModel

class ChunkMerger:
    @staticmethod
    def get_default_model():
        return ChunkMergeModel() \
            .setInputCol("entities") \
            .setOutputCol("merged_entities")





from sparknlp.annotator import *

class NGram:
    @staticmethod
    def get_default_model():
        return NGramGenerator() \
            .setInputCols(["token"]) \
            .setOutputCol("ngrams") \
            .setN(2) \
            .setEnableCumulative(False)




"""Collection of methods to substitute cols of licensed component results"""
import logging
logger = logging.getLogger('nlu')

def substitute_ner_internal_converter_cols(c, cols, nlu_identifier):
    """
    Fetched fields are:
    - entities@<storage_ref>_results
    - entities@<storage_ref>_<metadata>
        - entities@<storage_ref>_entity
        - entities@<storage_ref>_confidence
    """
    new_cols = {}
    new_base_name = 'entities' if nlu_identifier=='UNIQUE' else f'entities_{nlu_identifier}'
    for col in cols :
        if 'results'     in col     : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : new_cols[col] = f'{new_base_name}_embedding'
        elif 'meta' in col:
            if 'confidence' in col: new_cols[col]= f"{new_base_name}_confidence"
            elif 'entity' in     col: new_cols[col]= f"{new_base_name}_class"
            elif 'chunk' in     col: new_cols[col]= f"{new_base_name}_origin_chunk"
            elif 'sentence' in     col: new_cols[col]= f"{new_base_name}_origin_sentence"
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols





def substitute_chunk_resolution_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Resolution. For Resolution, some name will be infered, and entity_resolution_<name> will become the base name schema
all_k_results -> Sorted ResolverLabels in the top `alternatives` that match the distance `threshold`
all_k_resolutions -> Respective ResolverNormalized strings
all_k_distances -> Respective distance values after aggregation
all_k_wmd_distances -> Respective WMD distance values
all_k_tfidf_distances -> Respective TFIDF Cosinge distance values
all_k_jaccard_distances -> Respective Jaccard distance values
all_k_sorensen_distances -> Respective SorensenDice distance values
all_k_jaro_distances -> Respective JaroWinkler distance values
all_k_levenshtein_distances -> Respective Levenshtein distance values
all_k_confidences -> Respective normalized probabilities based in inverse distance values
target_text -> The actual searched string
resolved_text -> The top ResolverNormalized string
confidence -> Top probability
distance -> Top distance value
sentence -> Sentence index
chunk -> Chunk Index
token -> Token index
    """
    new_cols = {}
    new_base_name = f'entity_resolution' if nlu_identifier=='UNIQUE'  else f'entity_resolution_{nlu_identifier}'
    for col in cols :
        if '_results'      in col    and 'all_k' not in col :  new_cols[col] = f'{new_base_name}_code' # resolved code
        elif '_beginnings' in col     : new_cols[col]  = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col]  = f'{new_base_name}_end'
        elif '_types'      in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif '_embeddings' in col     : continue # omit , no data
        elif 'meta' in col:
            if   '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            elif 'all_k_aux_labels' in col  : new_cols[col] = f'{new_base_name}_k_aux_labels'  # maps to which sentence token comes from
            elif 'resolved_text' in col  : new_cols[col] = f'{new_base_name}' #The most likely resolution
            elif 'target_text' in col  : continue # Can be omitted, origin chunk basically, which will be included in the nerConverterInternal result
            elif 'all_k_confidences' in col  : new_cols[col] = f'{new_base_name}_k_confidences'  # confidences of the k resolutions
            elif 'confidence' in col  : new_cols[col] = f'{new_base_name}_confidence'
            elif 'all_k_results' in col  : new_cols[col] = f'{new_base_name}_k_results'
            elif 'all_k_distances' in col  : new_cols[col] = f'{new_base_name}_k_distances'
            elif 'all_k_resolutions' in col  : new_cols[col] = f'{new_base_name}_top_k'
            elif 'all_k_cosine_distances' in col  : new_cols[col] = f'{new_base_name}_k_cos_distances'
            elif 'all_k_wmd_distances' in col  : new_cols[col] = f'{new_base_name}_k_wmd_distances'
            elif 'all_k_tfidf_distances' in col  : new_cols[col] = f'{new_base_name}_k_tfidf_distances'
            elif 'all_k_jaccard_distances' in col  : new_cols[col] = f'{new_base_name}_k_jaccard_distances'
            elif 'all_k_sorensen_distances' in col  : new_cols[col] = f'{new_base_name}_k_sorensen_distances'
            elif 'all_k_jaro_distances' in col  : new_cols[col] = f'{new_base_name}_k_jaro_distances'
            elif 'all_k_levenshtein_distances' in col  : new_cols[col] = f'{new_base_name}_k_levenshtein_distances'
            elif 'distance' in col  : new_cols[col] = f'{new_base_name}_distance'
            elif 'chunk' in col  : continue # Omit, irreleant new_cols[col] = f'{new_base_name}_confidence'
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')

    return new_cols


def substitute_sentence_resolution_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Resolution. For Resolution, some name will be infered, and sentence_resolution_<name> will become the base name schema
all_k_results -> Sorted ResolverLabels in the top `alternatives` that match the distance `threshold`
all_k_resolutions -> Respective ResolverNormalized strings
all_k_distances -> Respective distance values after aggregation
all_k_wmd_distances -> Respective WMD distance values
all_k_tfidf_distances -> Respective TFIDF Cosinge distance values
all_k_jaccard_distances -> Respective Jaccard distance values
all_k_sorensen_distances -> Respective SorensenDice distance values
all_k_jaro_distances -> Respective JaroWinkler distance values
all_k_levenshtein_distances -> Respective Levenshtein distance values
all_k_confidences -> Respective normalized probabilities based in inverse distance values
target_text -> The actual searched string
resolved_text -> The top ResolverNormalized string
confidence -> Top probability
distance -> Top distance value
sentence -> Sentence index
chunk -> Chunk Index
token -> Token index
    """
    new_cols = {}
    new_base_name = f'sentence_resolution' if nlu_identifier=='UNIQUE' else f'sentence_resolution_{nlu_identifier}'
    for col in cols :
        if '_results'      in col    and 'all_k' not in col :  new_cols[col] = f'{new_base_name}_code' # resolved code
        elif '_beginnings' in col     : new_cols[col]  = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col]  = f'{new_base_name}_end'
        elif '_types'      in col          : continue # new_cols[col] = f'{new_base_name}_type'
        elif '_embeddings' in col     : continue # omit , no data
        elif 'meta' in col:
            if 'all_k_aux_labels' in col  : new_cols[col] = f'{new_base_name}_k_aux_labels'  # maps to which sentence token comes from
            elif 'resolved_text' in col  : new_cols[col] = f'{new_base_name}' #The most likely resolution
            elif 'target_text' in col  : continue # Can be omitted, origin chunk basically, which will be included in the nerConverterInternal result
            elif 'all_k_confidences' in col  : new_cols[col] = f'{new_base_name}_k_confidences'  # confidences of the k resolutions
            elif 'confidence' in col  : new_cols[col] = f'{new_base_name}_confidence'
            elif 'all_k_results' in col  : new_cols[col] = f'{new_base_name}_k_results'
            elif 'all_k_distances' in col  : new_cols[col] = f'{new_base_name}_k_distances'
            elif 'all_k_resolutions' in col  : new_cols[col] = f'{new_base_name}_top_k'
            elif 'all_k_cosine_distances' in col  : new_cols[col] = f'{new_base_name}_k_cos_distances'
            elif 'all_k_wmd_distances' in col  : new_cols[col] = f'{new_base_name}_k_wmd_distances'
            elif 'all_k_tfidf_distances' in col  : new_cols[col] = f'{new_base_name}_k_tfidf_distances'
            elif 'all_k_jaccard_distances' in col  : new_cols[col] = f'{new_base_name}_k_jaccard_distances'
            elif 'all_k_sorensen_distances' in col  : new_cols[col] = f'{new_base_name}_k_sorensen_distances'
            elif 'all_k_jaro_distances' in col  : new_cols[col] = f'{new_base_name}_k_jaro_distances'
            elif 'all_k_levenshtein_distances' in col  : new_cols[col] = f'{new_base_name}_k_levenshtein_distances'
            elif 'distance' in col  : new_cols[col] = f'{new_base_name}_distance'
            elif 'chunk' in col  : continue # Omit, irreleant new_cols[col] = f'{new_base_name}_confidence'
            elif   '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')

    return new_cols



def substitute_assertion_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for Assertion. For Assertion, some name will be infered, and assertion_<sub_field> defines the base name schema
    Assert should always be unique
    """
    new_cols = {}
    # c_name   = extract_nlu_identifier(c)
    new_base_name = f'assertion'# if is_unique else f'sentence_resolution_{c_name}'
    for col in cols :
        if '_results'      in col     :  new_cols[col] = f'{new_base_name}' # resolved code
        elif '_beginnings' in col     : new_cols[col]  = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col]  = f'{new_base_name}_end'
        elif '_types'      in col     : continue # new_cols[col] = f'{new_base_name}_type'
        elif '_embeddings' in col     : continue # omit , no data
        elif 'meta' in col:
            if   '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            elif 'chunk' in col : new_cols[col] = f'{new_base_name}_origin_chunk'  # maps to which sentence token comes from
            elif 'confidence' in col  : new_cols[col] = f'{new_base_name}' #The most likely resolution
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')

    return new_cols




def substitute_de_identification_cols(c, cols, is_unique=True):
    """
    Substitute col name for de-identification. For de-identification, some name will be infered, and de_identified_<sub_field> defines the base name schema
    de_identify should always be unique
    """
    new_cols = {}
    new_base_name = f'de_identified'#
    for col in cols :
        if '_results'      in col     :  new_cols[col] = f'{new_base_name}' # resolved code
        elif '_beginnings' in col     : new_cols[col]  = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col]  = f'{new_base_name}_end'
        elif '_types'      in col     : continue # new_cols[col] = f'{new_base_name}_type'
        elif '_embeddings' in col     : continue # omit , no data
        elif 'meta' in col:
            if   '_sentence' in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')

    return new_cols



def substitute_relation_cols(c, cols, nlu_identifier=True):
    """
    Substitute col name for de-identification. For de-identification, some name will be infered, and de_identified_<sub_field> defines the base name schema
    de_identify should always be unique

              metadata = Map(
                "entity1" -> relation.entity1,
                "entity2" -> relation.entity2,
                "entity1_begin" -> relation.entity1_begin.toString,
                "entity1_end" -> relation.entity1_end.toString,
                "entity2_begin" -> relation.entity2_begin.toString,
                "entity2_end" -> relation.entity2_end.toString,
                "chunk1" -> relation.chunk1,
                "chunk2" -> relation.chunk2,
                "confidence" -> result._2.toString
              ),

    """
    new_cols = {}
    new_base_name = f'relation' if nlu_identifier=='UNIQUE' else f'relation_{nlu_identifier}'
    for col in cols :
        if '_results'      in col     : new_cols[col]  = f'{new_base_name}' # resolved code
        elif '_beginnings' in col     : new_cols[col]  = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col]  = f'{new_base_name}_end'
        elif '_types'      in col     : continue # new_cols[col] = f'{new_base_name}_type'
        elif '_embeddings' in col     : continue # omit , no data
        elif 'meta' in col:
            if   '_sentence'       in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            elif   'entity1_begin' in col  : new_cols[col] = f'{new_base_name}_entity1_begin'  # maps to which sentence token comes from
            elif   'entity2_begin' in col  : new_cols[col] = f'{new_base_name}_entity2_begin'  # maps to which sentence token comes from
            elif   'entity1_end'   in col  : new_cols[col] = f'{new_base_name}_entity1_end'  # maps to which sentence token comes from
            elif   'entity2_end'   in col  : new_cols[col] = f'{new_base_name}_entity2_end'  # maps to which sentence token comes from
            elif   'confidence'    in col  : new_cols[col] = f'{new_base_name}_confidence'  # maps to which sentence token comes from
            elif   'entity1'       in col  : new_cols[col] = f'{new_base_name}_entity1_class'  # maps to which sentence token comes from
            elif   'entity2'       in col  : new_cols[col] = f'{new_base_name}_entity2_class'  # maps to which sentence token comes from
            elif   'chunk1'        in col  : new_cols[col] = f'{new_base_name}_entity1'  # maps to which sentence token comes from
            elif   'chunk2'        in col  : new_cols[col] = f'{new_base_name}_entity2'  # maps to which sentence token comes from
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    return new_cols


def substitute_drug_normalizer_cols(c, cols, is_unique=True):
    """
    Drug Norm is always unique
    Fetched fields are:
    - entities@<storage_ref>_results
    - entities@<storage_ref>_<metadata>
        - entities@<storage_ref>_entity
        - entities@<storage_ref>_confidence
    """
    new_cols = {}
    new_base_name = 'drug_norm'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'    in col     : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : continue # irrelevant  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types'      in col     : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if   '_sentence'       in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols



def substitute_context_parser_cols(c, cols, is_unique=True):
    """
    Drug Norm is always unique
    Fetched fields are:
    - entities@<storage_ref>_results
    - entities@<storage_ref>_<metadata>
        - entities@<storage_ref>_entity
        - entities@<storage_ref>_confidence
    """
    new_cols = {}
    new_base_name = 'context_match'# if is_unique else f'document_{nlu_identifier}'
    for col in cols :
        if '_results'    in col     : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : continue # irrelevant  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types'      in col     : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if   '_sentence'       in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            elif   'field'       in col  : new_cols[col] = f'{new_base_name}_field'  # maps to which sentence token comes from
            elif   'normalized'       in col  : new_cols[col] = f'{new_base_name}_normalized'  # maps to which sentence token comes from
            elif   'confidenceValue'       in col  : new_cols[col] = f'{new_base_name}_confidence'  # maps to which sentence token comes from
            elif   'hits'       in col  : new_cols[col] = f'{new_base_name}_hits'  # maps to which sentence token comes from

        else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
            # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols


def substitute_generic_classifier_parser_cols(c, cols, is_unique=True, nlu_identifier=''):
    """
    Drug Norm is always unique
    Fetched fields are:
    - entities@<storage_ref>_results
    - entities@<storage_ref>_<metadata>
        - entities@<storage_ref>_entity
        - entities@<storage_ref>_confidence
    """
    new_cols = {}
    new_base_name = 'generic_classifier' if is_unique else f'generic_classification_{nlu_identifier}'
    for col in cols :
        if '_results'    in col     : new_cols[col] = new_base_name
        elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
        elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
        elif '_embeddings' in col     : continue # irrelevant  new_cols[col] = f'{new_base_name}_embedding'
        elif '_types'      in col     : continue # new_cols[col] = f'{new_base_name}_type'
        elif 'meta' in col:
            if   '_sentence'       in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
            elif   'confidence'       in col  : new_cols[col] = f'{new_base_name}_confidence'  # maps to which sentence token comes from

        else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
        # new_cols[col]= f"{new_base_name}_confidence"
    return new_cols





"""
Verify all dependencies for NLU are requried
        # 1. Verfiy PYSPARK INSTALLED
        # 2. Verify SPARK-NLP INSTALLED
        # 3. Verify JAVA8 IS DEFAULT
        # FAILURE :
        # IF ST IMPORTABLE, TRY WRITE WARN MESSAGE TO STREAMLIT FROM NLU.LOAD()
        # IF ST NOT IMPORTABLE. JUST PRINT ERR MESSAGE
        # 4. ELSE LINK TO INSTALL NOTES https://nlu.johnsnowlabs.com/docs/en/install  and SLACK https://join.slack.com/t/spark-nlp/shared_invite/zt-lutct9gm-kuUazcyFKhuGY3_0AMkxqA



"""

# TODO
def verify_pyspark_installed(): pass

def verify_spark_nlp_installed(): pass

def verify_java8_installed_and_default(correct=''): pass

def verify_python_version(correct=''): pass


"""Define annotators for which names should be deducted using the mapped function"""
from nlu.pipe.col_substitution.col_substitution_HC import *
from nlu.pipe.col_substitution.col_substitution_OS import *

from sparknlp_jsl.annotator  import *
from sparknlp_jsl.base  import *


name_deductable_HC=[
    MedicalNerModel,
    NerConverterInternal,
    SentenceEntityResolverModel,
    ChunkEntityResolverModel,
    RelationExtractionModel,
    RelationExtractionDLModel,
    PosologyREModel,
    # Chunk2Token,
    ContextualParserModel, # todo
    DrugNormalizer,
    GenericClassifierModel,
    GenericClassifierApproach,
    FeaturesAssembler,
    # NerDisambiguatorModel
    # ChunkMergeModel
    # RENerChunksFilter
    NerOverwriter,
    ContextualParserModel, # todo
    ContextualParserApproach,

]



always_name_deductable_HC=[
    SentenceEntityResolverModel,
    ChunkEntityResolverModel,
    RelationExtractionModel,
    RelationExtractionDLModel,
    PosologyREModel,
    # Chunk2Token,
    ContextualParserModel, # todo
    ContextualParserApproach,
    # GenericClassifierModel,
    # NerDisambiguatorModel
    # ChunkMergeModel
    # RENerChunksFilter
    NerOverwriter,
    GenericClassifierModel,
    GenericClassifierApproach,
    DrugNormalizer,

]






__version__ = '3.1.1'
hard_offline_checks = False
def version(): return __version__
#
# from nlu.component_resolution import parse_language_from_nlu_ref, nlu_ref_to_component, \
#     construct_component_from_pipe_identifier


# if not check_pyspark_install(): raise Exception()
def try_import_pyspark_in_streamlit():
    """Try importing Pyspark or display warn message in streamlit"""
    try :
        import pyspark
        from pyspark.sql import SparkSession
    except:
        print("You need Pyspark installed to run NLU. Run <pip install pyspark==3.0.2>")

        try :
            import streamlit as st
            st.error("You need Pyspark, Sklearn, Pyplot, Pandas, Numpy installed to run this app. Run <pip install pyspark==3.0.2 sklearn pyplot numpy pandas>")
        except:
            return False
        return False
    return True
if not try_import_pyspark_in_streamlit() : raise  ImportError("You ned to install Pyspark")
st_cache_enabled = False
from typing import Optional

import nlu.utils.environment.env_utils as env_utils
import nlu.utils.environment.authentication as auth_utils
if not env_utils.check_python_version(): raise Exception()

import nlu
import logging
from nlu.spellbook import Spellbook
import warnings
from nlu.utils.environment.authentication import *

warnings.filterwarnings("ignore")

logger = logging.getLogger('nlu')
# logger.setLevel(logging.INFO)
logger.setLevel(logging.CRITICAL)
ch = logging.StreamHandler()
ch.setLevel(logging.CRITICAL)

logger.addHandler(ch)

# NLU Healthcare components
from nlu.components.assertion import Asserter
from nlu.components.resolution import Resolver
from nlu.components.relation import Relation
from nlu.components.deidentification import Deidentification

from nlu.components.embeddings.distil_bert.distilbert import DistilBert
from nlu.components.embeddings.roberta.roberta import Roberta
from nlu.components.embeddings.xlm.xlm import XLM



from nlu.components.utils.sentence_embeddings.spark_nlp_sentence_embedding import SparkNLPSentenceEmbeddings

from nlu import info
from nlu.info import ComponentInfo
from nlu.components import tokenizer, stemmer, spell_checker, normalizer, lemmatizer, embeddings, chunker, \
    embeddings_chunker
# Main components
from nlu.components.classifier import Classifier
from nlu.components.lemmatizer import Lemmatizer
from nlu.components.spell_checker import SpellChecker
from nlu.components.labeled_dependency_parser import LabeledDependencyParser as LabledDepParser
from nlu.components.unlabeled_dependency_parser import UnlabeledDependencyParser as UnlabledDepParser
from nlu.components.sentence_detector import NLUSentenceDetector

from nlu.components.dependency_untypeds.unlabeled_dependency_parser.unlabeled_dependency_parser import \
    UnlabeledDependencyParser
from nlu.components.dependency_typeds.labeled_dependency_parser.labeled_dependency_parser import \
    LabeledDependencyParser

# 0 Base internal Spark NLP structure required for all JSL components
from nlu.components.utils.document_assembler.spark_nlp_document_assembler import SparkNlpDocumentAssembler
from nlu.components.utils.ner_to_chunk_converter.ner_to_chunk_converter import NerToChunkConverter

# we cant call the embdding file "embeddings" because namespacing wont let us import the Embeddings class inside of it then
from nlu.components.embedding import Embeddings
from nlu.components.util import Util
from nlu.components.utils.ner_to_chunk_converter import ner_to_chunk_converter

# sentence
from nlu.components.sentence_detectors.pragmatic_sentence_detector.sentence_detector import PragmaticSentenceDetector
from nlu.components.sentence_detectors.deep_sentence_detector.deep_sentence_detector import SentenceDetectorDeep
# Embeddings
from nlu.components.embeddings.albert.spark_nlp_albert import SparkNLPAlbert
from nlu.components.embeddings.sentence_bert.BertSentenceEmbedding import BertSentence

from nlu.components.embeddings.bert.spark_nlp_bert import SparkNLPBert
from nlu.components.embeddings.elmo.spark_nlp_elmo import SparkNLPElmo
from nlu.components.embeddings.xlnet.spark_nlp_xlnet import SparkNLPXlnet
from nlu.components.embeddings.use.spark_nlp_use import SparkNLPUse
from nlu.components.embeddings.glove.glove import Glove

# classifiers
from nlu.components.classifiers.classifier_dl.classifier_dl import ClassifierDl
from nlu.components.classifiers.multi_classifier.multi_classifier import MultiClassifier
from nlu.components.classifiers.yake.yake import Yake
from nlu.components.classifiers.language_detector.language_detector import LanguageDetector
from nlu.components.classifiers.named_entity_recognizer_crf.ner_crf import NERDLCRF
from nlu.components.classifiers.ner.ner_dl import NERDL
from nlu.components.classifiers.sentiment_dl.sentiment_dl import SentimentDl
from nlu.components.classifiers.vivekn_sentiment.vivekn_sentiment_detector import ViveknSentiment
from nlu.components.classifiers.pos.part_of_speech_jsl import PartOfSpeechJsl

# matchers
from nlu.components.matchers.date_matcher.date_matcher import DateMatcher
from nlu.components.matchers.regex_matcher.regex_matcher import RegexMatcher
from nlu.components.matchers.text_matcher.text_matcher import TextMatcher

from nlu.components.matcher import Matcher

# token level operators
from nlu.components.tokenizer import Tokenizer
from nlu.components.lemmatizer import Lemmatizer
from nlu.components.stemmer import Stemmer
from nlu.components.normalizer import Normalizer
from nlu.components.stopwordscleaner import StopWordsCleaner as StopWordsCleaners
from nlu.components.stemmer import Stemmer as Stemmers
from nlu.components.stemmers.stemmer.spark_nlp_stemmer import SparkNLPStemmer
from nlu.components.normalizers.normalizer.spark_nlp_normalizer import SparkNLPNormalizer
from nlu.components.normalizers.document_normalizer.spark_nlp_document_normalizer import SparkNLPDocumentNormalizer

from nlu.components.lemmatizers.lemmatizer.spark_nlp_lemmatizer import SparkNLPLemmatizer
from nlu.components.stopwordscleaners.stopwordcleaner.nlustopwordcleaner import NLUStopWordcleaner
## spell
from nlu.components.spell_checkers.norvig_spell.norvig_spell_checker import NorvigSpellChecker
from nlu.components.spell_checkers.context_spell.context_spell_checker import ContextSpellChecker
from nlu.components.spell_checkers.symmetric_spell.symmetric_spell_checker import SymmetricSpellChecker
from nlu.components.tokenizers.default_tokenizer.default_tokenizer import DefaultTokenizer
from nlu.components.tokenizers.word_segmenter.word_segmenter import WordSegmenter

from nlu.components.chunkers.default_chunker.default_chunker import DefaultChunker
from nlu.components.embeddings_chunks.chunk_embedder.chunk_embedder import ChunkEmbedder
from nlu.components.chunkers.ngram.ngram import NGram

# sentence
from nlu.components.utils.sentence_detector.sentence_detector import SparkNLPSentenceDetector
from nlu.components.sentence_detector import NLUSentenceDetector

#seq2seq
from nlu.components.seq2seqs.marian.marian import Marian
from nlu.components.seq2seqs.t5.t5 import T5
from nlu.components.sequence2sequence import Seq2Seq

from nlu.pipe.pipeline import NLUPipeline
from nlu.pipe.utils.pipe_utils import PipeUtils
from nlu.pipe.pipe_logic import PipelineQueryVerifier
from nlu.pipe.component_resolution import *
global spark, all_components_info, nlu_package_location,authorized
is_authenticated=False
nlu_package_location = nlu.__file__[:-11]

spark_started = False
spark = None
authenticated = False
from nlu.info import AllComponentsInfo
from nlu.discovery import Discoverer
all_components_info = nlu.AllComponentsInfo()
discoverer = nlu.Discoverer()

import  json
import sparknlp
import os

def auth(SPARK_NLP_LICENSE_OR_JSON_PATH='/content/spark_nlp_for_healthcare.json',AWS_ACCESS_KEY_ID='',AWS_SECRET_ACCESS_KEY='',JSL_SECRET='', gpu=False):
    """ Authenticate enviroment for JSL Liscensed models. Installs NLP-Healthcare if not in enviroment detected
    Either provide path to spark_nlp_for_healthcare.json file as first param or manually enter them, SPARK_NLP_LICENSE_OR_JSON_PATH,AWS_ACCESS_KEY_ID,AWS_SECRET_ACCESS_KEY,JSL_SECRET .
    Set gpu=true if you want to enable GPU mode
    """
    if os.path.exists(SPARK_NLP_LICENSE_OR_JSON_PATH):
        with open(SPARK_NLP_LICENSE_OR_JSON_PATH) as json_file:
            j = json.load(json_file)
            auth_utils.get_authenticated_spark(j['SPARK_NLP_LICENSE'],  j['AWS_ACCESS_KEY_ID'],   j['AWS_SECRET_ACCESS_KEY'], j['SECRET'],gpu)
        return nlu
    if AWS_ACCESS_KEY_ID != '':
        auth_utils.get_authenticated_spark(SPARK_NLP_LICENSE_OR_JSON_PATH,AWS_ACCESS_KEY_ID,AWS_SECRET_ACCESS_KEY,JSL_SECRET, gpu)
    else : return nlu
    return nlu


def load_nlu_pipe_from_hdd(pipe_path,request)-> NLUPipeline:
    """Either there is a pipeline of models in the path or just one singular model.
    If it is a pipe,  load the pipe and return it.
    If it is a singular model, load it to the correct AnnotatorClass and NLU component and then generate pipeline for it
    """
    pipe = NLUPipeline()
    # if env_utils.is_running_in_databricks() :
    #     if pipe_path.startswith('/dbfs/') or pipe_path.startswith('dbfs/'):
    #         nlu_path = pipe_path
    #         if pipe_path.startswith('/dbfs/'):
    #             nlp_path =  pipe_path.replace('/dbfs','')
    #         else :
    #             nlp_path =  pipe_path.replace('dbfs','')
    #     else :
    #         nlu_path = 'dbfs/' + pipe_path
    #         if pipe_path.startswith('/') : nlp_path = pipe_path
    #         else : nlp_path = '/' + pipe_path
    nlu_ref=request# pipe_path
    if os.path.exists(pipe_path):
        if offline_utils.is_pipe(pipe_path):
            # language, nlp_ref, nlu_ref,path=None, is_licensed=False
            # todo deduct lang and if Licensed or not

            pipe_components = construct_component_from_pipe_identifier('en', nlu_ref, nlu_ref, pipe_path, False)
        elif offline_utils.is_model(pipe_path):
            c = offline_utils.verify_and_create_model(pipe_path)
            c.info.nlu_ref = nlu_ref
            pipe.add(c, nlu_ref, pretrained_pipe_component=True)
            return PipelineQueryVerifier.check_and_fix_nlu_pipeline(pipe)

        else :
            print(f"Could not load model in path {pipe_path}. Make sure the folder contains either a stages subfolder or a metadata subfolder.")
            raise ValueError
        for c in pipe_components: pipe.add(c, nlu_ref, pretrained_pipe_component=True)
        return pipe

    else :
        print(f"Could not load model in path {pipe_path}. Make sure the folder contains either a stages subfolder or a metadata subfolder.")
        raise ValueError


def enable_verbose()-> None:
    logger.setLevel(logging.INFO)
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    logger.addHandler(ch)

def disable_verbose()-> None:
    logger.setLevel(logging.ERROR)
    ch = logging.StreamHandler()
    ch.setLevel(logging.ERROR)
    logger.addHandler(ch)



def get_open_source_spark_context(gpu):
    if is_env_pyspark_2_3(): return sparknlp.start(spark23=True, gpu=gpu)
    if is_env_pyspark_2_4(): return sparknlp.start(spark24=True, gpu=gpu)
    if is_env_pyspark_3_0()  or is_env_pyspark_3_1(): return sparknlp.start(gpu=gpu)
    print(f"Current Spark version {get_pyspark_version()} not supported!")
    raise ValueError


def enable_streamlit_caching():
    # dynamically monkeypatch aka replace the nlu.load() method with the wrapped st.cache if streamlit_caching
    nlu.st_cache_enabled = True
    nlu.non_caching_load = load
    nlu.load =  wrap_with_st_cache_if_avaiable(nlu.load)

# def disable_streamlit_caching(): # WIP not working
#     if hasattr(nlu, 'non_caching_load') : nlu.load = nlu.non_caching_load
#     else : print("Could not disable caching.")


def wrap_with_st_cache_if_avaiable(f):
    """Wrap function with ST cache method if streamlit is importable"""
    try :
        import streamlit as st
        logger.info("Using streamlit cache for load")
        return st.cache(f, allow_output_mutation=True, show_spinner=False)
    except :
        logger.exception("Could not import streamlit and apply caching")
        print("You need streamlit to run use this method")
        return f


def enable_hard_offline_checks():nlu.hard_offline_checks = True
def disable_hard_offline_checks():nlu.hard_offline_checks = False



def load(request:str ='from_disk', path:Optional[str]=None,verbose:bool=False, gpu:bool=False, streamlit_caching:bool=False)->NLUPipeline :
    '''
    Load either a prebuild pipeline or a set of components identified by a whitespace seperated list of components
    You must call nlu.auth() BEFORE calling nlu.load() to access licensed models.
    If you did not call nlu.auth() but did call nlu.load() you must RESTART your Python Process and call nlu.auth().
    You cannot authorize once nlu.load() is called because of Spark Context.
    :param verbose:
    :param path: If path is not None, the model/pipe for the NLU reference will be loaded from the path. Useful for offline mode. Currently only loading entire NLU pipelines is supported, but not loading singular pipes
    :param request: A NLU model/pipeline/component reference
    :param version_checks: Wether to check if Pyspark is properly installed and if the Pyspark version is correct for the NLU version. If set to False, these tests will be skipped
    :return: returns a non fitted nlu pipeline object
    '''
    if streamlit_caching and not nlu.st_cache_enabled :
        enable_streamlit_caching()
        return nlu.load(request, path,verbose, gpu, streamlit_caching)
    global is_authenticated
    is_authenticated = True
    auth(gpu=gpu) # check if secets are in default loc, if yes load them and create licensed context automatically
    spark = get_open_source_spark_context(gpu)
    spark.catalog.clearCache()
    if verbose:enable_verbose()
    else: disable_verbose()


    if path != None :
        logger.info(f'Trying to load nlu pipeline from local hard drive, located at {path}')
        pipe = PipelineQueryVerifier.check_and_fix_nlu_pipeline(load_nlu_pipe_from_hdd(path,request))
        pipe.nlu_ref = request
        return pipe
    components_requested = request.split(' ')
    pipe = NLUPipeline()
    language = parse_language_from_nlu_ref(request)
    pipe.lang=language
    pipe.nlu_ref = request

    try :
        for nlu_ref in components_requested:
            nlu_ref.replace(' ', '')
            # component = component.lower()
            if nlu_ref == '': continue
            nlu_component = nlu_ref_to_component(nlu_ref, authenticated=is_authenticated)
            # if we get a list of components, then the NLU reference is a pipeline, we do not need to check order
            if type(nlu_component) == type([]):
                # lists are parsed down to multiple components
                for c in nlu_component: pipe.add(c, nlu_ref, pretrained_pipe_component=True)
            else:
                pipe.add(nlu_component, nlu_ref)
        pipe = PipelineQueryVerifier.check_and_fix_nlu_pipeline(pipe)
        pipe.nlu_ref = request
        for c in pipe.components :
            if c.info.license == 'licensed' : pipe.has_licensed_components=True
        return pipe

    except:
        import sys
        if verbose:
            e = sys.exc_info()
            print(e[0])
            print(e[1])

        # 1. Verfiy PYSPARK INSTALLED
        # 2. Verify SPARK-NLP INSTALLED
        # 3. Verify JAVA8 IS DEFAULT
        # FAILURE :
        # IF ST IMPORTABLE, TRY WRITE WARN MESSAGE TO STREAMLIT FROM NLU.LOAD()
        # IF ST NOT IMPORTABLE. JUST PRINT ERR MESSAGE
        # 4. ELSE LINK TO INSTALL NOTES https://nlu.johnsnowlabs.com/docs/en/install  and SLACK https://join.slack.com/t/spark-nlp/shared_invite/zt-lutct9gm-kuUazcyFKhuGY3_0AMkxqA
        raise Exception("Something went wrong during loading and fitting the pipe. Check the other prints for more information and also verbose mode. Did you use a correct model reference?")



class NluError:
    def __init__(self):
        self.has_trainable_components = False
    @staticmethod
    def predict(text, output_level='error', positions='error', metadata='error', drop_irrelevant_cols=False):
        print('The NLU components could not be properly created. Please check previous print messages and Verbose mode for further info')
    @staticmethod
    def print_info(): print("Sorry something went wrong when building the pipeline. Please check verbose mode and your NLU reference.")



#Discovery
def print_all_languages():
    ''' Print all languages which are available in NLU Spark NLP pointer '''
    discoverer.print_all_languages()
def print_all_nlu_components_for_lang(lang='en', c_type='classifier'):
    '''Print all NLU components available for a language Spark NLP pointer'''
    discoverer.print_all_nlu_components_for_lang(lang, c_type)
def print_components(lang='', action=''):
    '''Print every single NLU reference for models and pipeliens and their Spark NLP pointer
    :param lang: Language requirements for the components filterd. See nlu.languages() for supported languages
    :param action: Components that will be filterd.'''
    discoverer.print_components(lang,action)
def print_component_types():
    ''' Prints all unique component types in NLU'''
    discoverer.print_component_types()
def print_all_model_kinds_for_action(action):
    discoverer.print_all_model_kinds_for_action(action)
def print_all_model_kinds_for_action_and_lang(lang, action):
    discoverer.print_all_model_kinds_for_action_and_lang(lang,action)
def print_trainable_components():
    '''Print every trainable Algorithm/Model'''
    discoverer.print_trainable_components()

def get_components(m_type='', include_pipes=False, lang='', licensed=False, get_all=False):
    return discoverer.get_components(m_type, include_pipes, lang, licensed, get_all)



from sparknlp.annotator import *

class SparkNLPDocumentNormalizer:
    @staticmethod
    def get_default_model():
        return DocumentNormalizer() \
            .setInputCols(["document"]) \
            .setOutputCol("normalized_document")\
            .setAction('clean')




"""
Map Healthcare annotators to output level
"""
from sparknlp_jsl.annotator  import *
HC_anno2output_level = {
    'document': [
        RENerChunksFilter,
        DrugNormalizer,

                 ],
    'sentence': [
        SentenceEntityResolverModel

    ],
    'chunk': [
        NerOverwriter,
        NerDisambiguatorModel,
        ChunkMergeModel,
        NerConverterInternal,
        ChunkEntityResolverModel,
        DeIdentificationModel,
        AssertionDLModel,
        AssertionLogRegModel,
        ContextualParserApproach,
        ContextualParserModel


    ],
    'token': [
        MedicalNerModel,

               ],
    'input_dependent': [
        GenericClassifierModel,
                        ],
    'sub_token': [DrugNormalizer,Chunk2Token],
    'leveless': [
        ContextualParserModel,
        ContextualParserModel,

        ContextualParserModel,

    ],
    'relation':[
        RelationExtractionModel,
        PosologyREModel,
        RelationExtractionDLModel,
    ]

}




from nlu.pipe.pipe_components import SparkNLUComponent


class UnlabeledDependencyParser(SparkNLUComponent):

    def __init__(self, annotator_class='unlabeled_dependency_parser', language='en', component_type='dependency_untyped', get_default = True, nlp_ref='', nlu_ref ='', model=None,loaded_from_pretrained_pipe=False,is_licensed=False):

        if model != None :self.model = model
        elif 'dep' in annotator_class or 'dep.untyped' in annotator_class or annotator_class== 'unlabeled_dependency_parser':
            from nlu.components.dependency_untypeds.unlabeled_dependency_parser.unlabeled_dependency_parser import UnlabeledDependencyParser
            if get_default : self.model = UnlabeledDependencyParser.get_default_model()
            else : self.model = UnlabeledDependencyParser.get_pretrained_model(nlp_ref, language)


        SparkNLUComponent.__init__(self, annotator_class, component_type, nlu_ref, nlp_ref, language,loaded_from_pretrained_pipe , is_licensed)



import pandas as pd
import json
p = '/home/ckl/Documents/freelance/jsl/nlu/nlu4realgit2/tmp/find_unused_models_in_s3/metadata.csv'
df=pd.read_csv(p)
import nlu
# CLEAN RESULT REGEX
# .\d.\d.\d.\d.\d.\d.\d,True,\d*-\d*-\d*T\d*:\d*:\d*.\d*Z,True,\D*,
unused = []
unused_idx =[]
for idx,row in df.iterrows():
    if row['language'] in nlu.Spellbook.pretrained_pipe_references.keys():
        if row['name'] in nlu.Spellbook.pretrained_pipe_references[row['language']].values(): continue

    if row['language'] in nlu.Spellbook.pretrained_models_references.keys():
        if row['name'] in nlu.Spellbook.pretrained_models_references[row['language']].values():
            continue

    unused.append(row['name'] + ' ->' + row['language'] + ' ->' + row['time'] + '\n')
    unused_idx.append(idx)

# for nlp_ref in unused : print(nlp_ref)
print('NUM UNSUED:\n', '\n'.join(unused))
#
# print('hi')
# # 1. translate_[from]_[to] (pipe)
# #  2. opus_mt_[from]_[to]
df['nlu_ref'] = ''

opus_idx = df.name.str.contains('opus_mt')
opus_idx = opus_idx[opus_idx==True].index
cand = df.iloc[opus_idx].name.str.split('_')
f = lambda x : 'xx.'+x[3] +'.marian.translate_to.' + x[2]
nlu_opus_refs = cand.map(f)
df.nlu_ref.iloc[opus_idx] = nlu_opus_refs[opus_idx]



translate_idx = df.name.str.contains('translate')
translate_idx = translate_idx[translate_idx==True].index
cand = df.iloc[translate_idx].name.str.split('_')
f = lambda x : 'xx.'+x[2] +'.translate_to.' + x[1]
nlu_translate_refs = cand.map(f)
# df.iloc[translate_idx ]['nlu_ref'] = nlu_translate_refs[translate_idx]
df.nlu_ref.iloc[translate_idx] = nlu_translate_refs[translate_idx]

print('writing the idnx')
print(unused_idx)
df.iloc[unused_idx][['nlu_ref','name']].to_csv('with_3.1_refs_clean.csv', index=False,)
# print(df)
# print(df)
# import json
# df.index = df.name
# # Create OPUS json NLU references. Put these into a JSON formatter, then add to namespace!
# j =df.iloc[opus_idx].to_json()
# d = json.loads(j)
#
# k = d['nlu_ref'].keys()
# v = d['nlu_ref'].values()
# d = dict(zip(v,k))
# print('JSON OPUS:')
# print(d)
#
#
#
# j =df.iloc[translate_idx].to_json()
# d = json.loads(j)
# k = d['nlu_ref'].keys()
# v = d['nlu_ref'].values()
# d = dict(zip(v,k))
# print('JSON TRANSLTE:')
# print(d)
#

## WRITE EVERY translate and opus ref to NLU!
# Marian pipes as default, i.e 'en.translate_to.fr' is default translate pip
# en.marian_to.de  is the model specific selector

# TODO WRITE NEW NLU REFS
# IDEA: the base refs are `en.translate_to.german` i.e. `<lang>.translate_to.<lang>
# add them all into the `xx` namespace
# when parsing, check if reference contains `translate_to`, if yes then prefix it with `xx` before further parsing
# Also set lang to xx since the sparknlp ref is actually also xx




import unittest
from nlu import *
class TestCyber(unittest.TestCase):


    def test_cyber_model(self):
        import pandas as pd
        pipe = nlu.load('cyberbullying',verbose=True)
        df = pipe.predict(['Peter love pancaces. I hate Mondays', 'I love Fridays'], output_level='token',drop_irrelevant_cols=False, metadata=True, )
        for c in df.columns: print(df[c])

if __name__ == '__main__':
    unittest.main()




import os,sys
import logging
logger = logging.getLogger('nlu')

def get_pyspark_version():
    import pyspark
    return pyspark.version.__version__

def get_pyspark_major_and_minor():
    import pyspark
    return pyspark.version.__version__.split('.')[:2]


def is_env_pyspark_3_1():
    v = get_pyspark_major_and_minor()
    if v[0] == '3' and v[1] == '1' : return True
    return False

def is_env_pyspark_3_0():
    v = get_pyspark_major_and_minor()
    if v[0] == '3' and v[1] == '0' : return True
    return False


def is_env_pyspark_2_4():
    v = get_pyspark_major_and_minor()
    if v[0] == '2' and v[1] == '4' : return True
    return False

def is_env_pyspark_2_3():
    v = get_pyspark_major_and_minor()
    if v[0] == '2' and v[1] == '3' : return True
    return False


def check_pyspark_install():
    try :
        from pyspark.sql import SparkSession
        try :
            import sparknlp
            v = sparknlp.start().version
            spark_major = int(v.split('.')[0])
            if spark_major >= 3 :
                raise Exception()
        except :
            print(f"Detected pyspark version={v} Which is >=3.X\nPlease run '!pip install pyspark==2.4.7' or install any pyspark>=2.4.0 and pyspark<3")
            # print(f"Or set nlu.load(version_checks=False). We disadvise from doing so, until Pyspark >=3 is officially supported in 2021.")
            return False
    except :
        print("No Pyspark installed!\nPlease run '!pip install pyspark==2.4.7' or install any pyspark>=2.4.0 with pyspark<3")
        return False
    return True

def check_python_version():
    if float(sys.version[:3]) >= 3.9:
        print("Please use a Python version with version number SMALLER than 3.9")
        print("Python versions equal or higher 3.9 are currently NOT SUPPORTED by NLU")
        return False
    return True

def is_running_in_databricks():
    """ Check if the currently running Python Process is running in Databricks or not
     If any Enviroment Variable name contains 'DATABRICKS' this will return True, otherwise False"""
    for k in os.environ.keys() :
        if 'DATABRICKS' in k :
            return True
    return False

def install_and_import_package(pkg_name,version='', import_name=''):
    """ Install Spark-NLP-Healthcare PyPI Package in current enviroment if it cannot be imported and liscense provided"""
    import importlib
    try:
        if import_name == '' : importlib.import_module(pkg_name)
        else: importlib.import_module(import_name)
    except ImportError:
        import pip
        if version == '':
            logger.info(f"{pkg_name} could not be imported. Running 'pip install {pkg_name}'...")
        else :
            logger.info(f"{pkg_name} could not be imported. Running 'pip install {pkg_name}=={version}'...")
        pip_major_version = int(pip.__version__.split('.')[0])
        if pip_major_version in [10, 18, 19,20]:
            # for these versions pip module does not support installing, we install via OS command straight into pip module
            py_path = sys.executable
            if version == '':
                os.system(f'{py_path} -m pip install {pkg_name}')
            else :
                os.system(f'{py_path} -m pip install {pkg_name}=={version}')
        else:
            if version == '':
                pip.main(['install', f'{pkg_name}'])
            else :
                pip.main(['install', f'{pkg_name}=={version}'])
    finally:

        import site
        from importlib import reload
        reload(site)
        # import name is not always the same name as pkg_name we want to import, so it must be specified via import name
        if import_name != '' : globals()[import_name] = importlib.import_module(import_name)
        else :globals()[pkg_name] = importlib.import_module(pkg_name)



"""Resolve output level of pipeline and components"""
from  typing import List, Dict
import logging
logger = logging.getLogger('nlu')
from sparknlp.base import *
from sparknlp.annotator import *
from nlu.pipe.col_substitution.col_name_substitution_utils import  ColSubstitutionUtils

"""Component and Column Level logic operations and utils"""
class OutputLevelUtils():
    levels = {
        'token': ['token', 'pos', 'ner', 'lemma', 'lem', 'stem', 'stemm', 'word_embeddings', 'named_entity',
                  'entity', 'dependency',
                  'labeled_dependency', 'dep', 'dep.untyped', 'dep.typed'],
        'sentence': ['sentence', 'sentence_embeddings', ] + ['sentiment', 'classifer', 'category'],
        'chunk': ['chunk', 'embeddings_chunk', 'chunk_embeddings'],
        'document': ['document', 'language'],
        'embedding_level': []
        # ['sentiment', 'classifer'] # WIP, wait for Spark NLP Getter/Setter fixes to implement this properly
        # embedding level  annotators output levels depend on the level of the embeddings they are fed. If we have Doc/Chunk/Word/Sentence embeddings, those annotators output at the same level.

    }
    annotator_levels_approach_based = {
        'document': [DocumentAssembler, Chunk2Doc,
                     YakeModel,
                     ],
        'sentence': [SentenceDetector, SentenceDetectorDLApproach, ],
        'chunk': [Chunker, ChunkEmbeddings,  ChunkTokenizer, Token2Chunk, TokenAssembler,
                  NerConverter, Doc2Chunk,NGramGenerator],
        'token': [ NerCrfApproach, NerDLApproach,
                   PerceptronApproach,
                   Stemmer,
                   ContextSpellCheckerApproach,
                   WordSegmenterApproach,
                   Lemmatizer,LemmatizerModel, TypedDependencyParserApproach, DependencyParserApproach,
                   Tokenizer, RegexTokenizer, RecursiveTokenizer
            , DateMatcher, TextMatcher, BigTextMatcher, MultiDateMatcher,
                   WordSegmenterApproach
                   ],
        # 'sub_token': [StopWordsCleaner, DateMatcher, TextMatcher, BigTextMatcher, MultiDateMatcher],
        # these can be document or sentence
        'input_dependent': [ViveknSentimentApproach, SentimentDLApproach, ClassifierDLApproach,
                            LanguageDetectorDL,
                            MultiClassifierDLApproach,  SentenceEmbeddings, NorvigSweetingApproach,
                            ],
        'multi' : [MultiClassifierDLApproach,  SentenceEmbeddings, NorvigSweetingApproach,]
        # 'unclassified': [Yake, Ngram]
    }
    annotator_levels_model_based = {
        'document': [],
        'sentence': [SentenceDetectorDLModel, ],
        'chunk': [ChunkTokenizerModel, ChunkTokenizerModel,  ],
        'token': [ContextSpellCheckerModel, AlbertEmbeddings, BertEmbeddings, ElmoEmbeddings, WordEmbeddings,
                  XlnetEmbeddings, WordEmbeddingsModel,
                  # NER models are token level, they give IOB predictions and cofidences for EVERY token!
                  NerDLModel, NerCrfModel, PerceptronModel, SymmetricDeleteModel, NorvigSweetingModel,
                  ContextSpellCheckerModel,
                  TypedDependencyParserModel, DependencyParserModel,
                  RecursiveTokenizerModel,
                  TextMatcherModel, BigTextMatcherModel, RegexMatcherModel,
                  WordSegmenterModel, TokenizerModel,
                  XlmRoBertaEmbeddings, RoBertaEmbeddings, DistilBertEmbeddings,
                  ],
        # 'sub_token': [TextMatcherModel, BigTextMatcherModel, RegexMatcherModel, ],
        # sub token is when annotator is token based but some tokens may be missing since dropped/cleaned
        'sub_token' : [
            StopWordsCleaner, NormalizerModel

        ] ,
        'input_dependent': [BertSentenceEmbeddings, UniversalSentenceEncoder, ViveknSentimentModel,
                            SentimentDLModel,  ClassifierDLModel,
                            MarianTransformer,T5Transformer,
                            XlmRoBertaEmbeddings, RoBertaEmbeddings, DistilBertEmbeddings,

                            ],
        'multi' : [MultiClassifierDLModel, MultiClassifierDLModel,]
    }
    all_embeddings = {
        'token' : [AlbertEmbeddings, BertEmbeddings, ElmoEmbeddings, WordEmbeddings,
                   XlnetEmbeddings,WordEmbeddingsModel],
        'input_dependent' : [SentenceEmbeddings, UniversalSentenceEncoder,BertSentenceEmbeddings]

    }

    @staticmethod
    def infer_output_level(pipe):
        '''
        This function checks the LAST  component of the NLU pipeline and infers
        and infers from that the output level via checking the components info.
        It sets the output level of the pipe accordingly
        param sdf : Spark dataframe after transformations
        '''
        if pipe.output_level == '' :
            # Loop in reverse over pipe and get first non util/sentence_detecotr/tokenizer/doc_assember. If there is non, take last
            bad_types = [ 'util','document','sentence']
            bad_names = ['token']
            for c in pipe.components[::-1]:
                if any (t in  c.info.type for t in bad_types) : continue
                if any (n in  c.info.name for n in bad_names) : continue
                pipe.output_level = OutputLevelUtils.resolve_component_to_output_level(pipe,c)
                logger.info(f'Inferred and set output level of pipeline to {pipe.output_level}', )
                break
            # Voodo Normalizer bug that does not happen in debugger bugfix
            if pipe.output_level == None  or pipe.output_level == '': pipe.output_level = 'document'
            logger.info(f'Inferred and set output level of pipeline to {pipe.output_level}' )

        else : return

    @staticmethod
    def get_output_level_of_embeddings_provider(pipe, field_type, field_name):
        '''
        This function will go through all components to find the component which  generate @component_output_column_name.
        Then it will go gain through all components to find the component, from which @component_output_column_name is taking its inputs
        Then it will return the type of the provider component. This result isused to resolve the output level of the component that depends on the inpit for the output level
        :param field_type: The type of the field we want to resolve the input level for
        :param field_name: The name of the field we want to resolve the input level for

        :return:
        '''
        # find the component. Column output name should be unique
        component_inputs = []
        for component in pipe.components:
            if field_name == component.info.name:
                component_inputs = component.info.spark_input_column_names

        # get the embedding feature name
        target_output_component = ''
        for input_name in component_inputs:
            if 'embed' in input_name: target_output_component = input_name

        # get the model that outputs that feature
        for component in pipe.components:
            component_outputs = component.info.spark_output_column_names
            for input_name in component_outputs:
                if target_output_component == input_name:
                    # this is the component that feeds into the component we are trying to resolve the output  level for.
                    # That is so, because the output of this component matches the input of the component we are resolving
                    return pipe.resolve_type_to_output_level(component.info.type)
    @staticmethod
    def resolve_type_to_output_level(pipe, field_type, field_name):
        '''
        This checks the levels dict for what the output level is for the input annotator type.
        If the annotator type depends on the embedding level, we need further checking.
        @ param field_type : type of the spark field
        @ param name : name of thhe spark field
        @ return : String, which corrosponds to the output level of this Component.
        '''
        logger.info('Resolving output level for field_type=%s and field_name=%s', field_type, field_name)
        if field_name == 'sentence':
            logger.info('Resolved output level for field_type=%s and field_name=%s to Sentence level', field_type,
                        field_name)
            return 'sentence'
        if field_type in pipe.levels['token']:
            logger.info('Resolved output level for field_type=%s and field_name=%s to Token level ', field_type,
                        field_name)
            return 'token'
        if field_type in pipe.levels['sentence']:
            logger.info('Resolved output level for field_type=%s and field_name=%s to sentence level', field_type,
                        field_name)
            return 'sentence'
        if field_type in pipe.levels['chunk']:
            logger.info('Resolved output level for field_type=%s and field_name=%s to Chunk level ', field_type,
                        field_name)
            return 'chunk'
        if field_type in pipe.levels['document']:
            logger.info('Resolved output level for field_type=%s and field_name=%s to document level', field_type,
                        field_name)
            return 'document'
        if field_type in pipe.levels['embedding_level']:
            logger.info('Resolved output level for field_type=%s and field_name=%s to embeddings level', field_type,
                        field_name)
            return pipe.get_output_level_of_embeddings_provider(field_type, field_name)  # recursive resolution
    @staticmethod
    def resolve_input_dependent_component_to_output_level(pipe, component):
        '''
        For a given NLU component  which is input dependent , resolve its output level by checking if it's input stem from document or sentence based annotators
        :param component:  to resolve
        :return: resolve component
        '''
        # (1.) A classifier, which is using sentence/document. We just check input cols

        if 'document' in component.info.spark_input_column_names :  return 'document'
        if 'sentence' in component.info.spark_input_column_names :  return 'sentence'

        # (2.) A classifier, which is using sentence/doc embeddings.
        # We iterate over the pipe and check which Embed component is feeding the classifier and what the input that embed annotator is (sent or doc)
        for c in pipe.components:
            # check if c is of sentence embedding class  which is always input dependent
            if any ( isinstance(c.model, e ) for e in OutputLevelUtils.all_embeddings['input_dependent']  ) :
                if 'document' in c.info.spark_input_column_names :  return 'document'
                if 'sentence' in c.info.spark_input_column_names :  return 'sentence'
    @staticmethod
    def resolve_component_to_output_level(pipe,component):
        '''
        For a given NLU component, resolve its output level, by checking annotator_levels dicts for approaches and models
        If output level is input dependent, resolve_input_dependent_component_to_output_level will resolve it
        :param component:  to resolve
        :return: resolve component
        '''
        for level in OutputLevelUtils.annotator_levels_model_based.keys():
            for t in OutputLevelUtils.annotator_levels_model_based[level]:
                if isinstance(component.model,t) :
                    if level == 'input_dependent' : return OutputLevelUtils.resolve_input_dependent_component_to_output_level(pipe,component)
                    else : return level
        for level in OutputLevelUtils.annotator_levels_approach_based.keys():
            for t in OutputLevelUtils.annotator_levels_approach_based[level]:
                if isinstance(component.model,t) :
                    if level == 'input_dependent' : return OutputLevelUtils.resolve_input_dependent_component_to_output_level(pipe,component)
                    else : return level
        if pipe.has_licensed_components:
            from nlu.pipe.extractors.output_level_HC_map import HC_anno2output_level
            for level in HC_anno2output_level.keys():
                for t in HC_anno2output_level[level]:
                    if isinstance(component.model,t) :
                        if level == 'input_dependent' : return OutputLevelUtils.resolve_input_dependent_component_to_output_level(pipe,component)
                        else : return level
    @staticmethod
    def get_output_level_mappings(pipe,df,anno_2_ex_config):
        """Get a dict where key=spark_colname and val=output_level, inferred from processed dataframe and pipe that is currently running"""
        output_level_map = {}
        same_output_level_map = {}
        not_same_output_level_map = {}
        for c in pipe.components:
            generated_cols = ColSubstitutionUtils.get_final_output_cols_of_component(c,df,anno_2_ex_config)
            output_level = OutputLevelUtils.resolve_component_to_output_level(pipe,c)
            if output_level == pipe.output_level :
                for g_c in generated_cols: same_output_level_map[g_c]=output_level
            else :
                for g_c in generated_cols: not_same_output_level_map[g_c]=output_level
            for g_c in generated_cols: output_level_map[g_c]=output_level
        return output_level_map,same_output_level_map,not_same_output_level_map
        # return {c.info.outputs[0] :OutputLevelUtils.resolve_component_to_output_level(pipe,c)  for c in pipe.components}
    @staticmethod
    def get_cols_at_same_output_level(pipe,df,anno_2_ex_config,col2output_level:Dict[str,str])->List[str]:
        """Get List of cols which are at same output level as the pipe is currently configured to"""
        # return [c.info.outputs[0]  for c in pipe.components if pipe.resolve_component_to_output_level(c) == pipe.output_level ]
        same_output_level_cols = []
        for c in pipe.components:
            if col2output_level[c.info.outputs[0]] == pipe.output_level:
                same_output_level_cols + ColSubstitutionUtils.get_final_output_cols_of_component(c,df,anno_2_ex_config)
        return same_output_level_cols
    @staticmethod
    def get_cols_not_at_same_output_level(pipe,df,anno_2_ex_config,col2output_level:Dict[str,str])->List[str]:
        """Get List of cols which are not at same output level as the pipe is currently configured to"""
        # return [c.info.outputs[0]  for c in pipe.components if not pipe.resolve_component_to_output_level(c) == pipe.output_level ]
        return [c.info.outputs[0]  for c in pipe.components if not col2output_level[c.info.outputs[0]] == pipe.output_level ]



    @staticmethod
    def get_output_level_mapping_by_component(pipe)->Dict[str,str]:
        """Get a dict where key=colname and val=output_level, inferred from processed dataframe and pipe that is currently running"""
        return {c :OutputLevelUtils.resolve_component_to_output_level(pipe,c)  for c in pipe.components}




import nlu
from nlu.discovery import Discoverer
from nlu.pipe.utils.storage_ref_utils import StorageRefUtils
from typing import List, Tuple, Optional, Dict, Union
import streamlit as st
from nlu.utils.modelhub.modelhub_utils import ModelHubUtils
import numpy as np
import pandas as pd
from nlu.pipe.viz.streamlit_viz.streamlit_utils_OS import StreamlitUtilsOS
from nlu.pipe.viz.streamlit_viz.gen_streamlit_code import get_code_for_viz
from nlu.pipe.viz.streamlit_viz.styles import _set_block_container_style
import random
from nlu.pipe.viz.streamlit_viz.streamlit_viz_tracker import StreamlitVizTracker
from nlu.pipe.viz.streamlit_viz.viz_building_blocks.dep_tree import DepTreeStreamlitBlock
from nlu.pipe.viz.streamlit_viz.viz_building_blocks.classifier import ClassifierStreamlitBlock
from nlu.pipe.viz.streamlit_viz.viz_building_blocks.token_features import TokenFeaturesStreamlitBlock
from nlu.pipe.viz.streamlit_viz.viz_building_blocks.ner import NERStreamlitBlock
from nlu.pipe.viz.streamlit_viz.viz_building_blocks.word_similarity import WordSimilarityStreamlitBlock
from nlu.pipe.viz.streamlit_viz.viz_building_blocks.word_embedding_manifold import WordEmbeddingManifoldStreamlitBlock



class StreamlitVizBlockHandler():
    """Internal API to access any on the Streamlit building blocks. This is part of the Controller of the MVC pattern"""
    @staticmethod
    def viz_streamlit_dashboard(
            pipe,
            # Base Params
            text:Union[str, List[str], pd.DataFrame, pd.Series],
            model_selection:List[str]=[],
            # NER PARAMS
            # default_ner_model2viz:Union[str, List[str]] = 'en.ner.onto.electra.base',
            # SIMILARITY PARAMS
            similarity_texts:Tuple[str,str]= ('I love NLU <3', 'I love Streamlit <3'),
            title:str = 'NLU ❤️ Streamlit - Prototype your NLP startup in 0 lines of code' ,
            sub_title:str = 'Play with over 1000+ scalable enterprise NLP models',
            side_info:str = None,
            # UI PARAMS
            visualizers:List[str] = ( "dependency_tree", "ner",  "similarity", "token_features", 'classification','manifold'),
            show_models_info:bool = True,
            show_model_select:bool = False,
            show_viz_selection:bool = False,
            show_logo:bool=True,
            set_wide_layout_CSS:bool=True,
            show_code_snippets:bool=False,
            model_select_position:str = 'side' , # main or side
            display_infos:bool=True,
            key:str = "NLU_streamlit",
            display_footer :bool =  True ,
            num_similarity_cols:int=2,

            # NEW PARAMS
            # MANIfold
            num_manifold_cols:int=3,
            manifold_algos:List[str]=('TSNE'),

            # SIMY
            similarity_algos:List[str]=('COSINE'),
    )-> None:
        """Visualize either individual building blocks for streamlit or a full UI to experiment and explore models with"""
        StreamlitVizTracker.footer_displayed = not display_footer
        if set_wide_layout_CSS : _set_block_container_style()
        if title: st.title(title)
        if sub_title: st.subheader(sub_title)
        if show_logo :StreamlitVizTracker.show_logo()
        if side_info : st.sidebar.markdown(side_info)
        text    = st.text_area("Enter text you want to visualize below", text, key=key)
        ner_model_2_viz     = pipe.nlu_ref
        if show_model_select :
            show_code_snippets = st.sidebar.checkbox('Generate code snippets', value=show_code_snippets)
            if model_selection == [] : model_selection = Discoverer.get_components('ner',include_pipes=True)
            model_selection.sort()
            if model_select_position == 'side':
                if pipe.nlu_ref.split(' ')[0] in  model_selection :
                    ner_model_2_viz = st.sidebar.selectbox("Select a NER model.",model_selection,index=model_selection.index(pipe.nlu_ref.split(' ')[0]))
                else :
                    ner_model_2_viz = st.sidebar.selectbox("Select a NER model.",model_selection,index=model_selection.index('en.ner'))
            else :
                if pipe.nlu_ref.split(' ')[0] in  model_selection :
                    ner_model_2_viz = st.selectbox("Select a NER model",model_selection,index=model_selection.index(pipe.nlu_ref.split(' ')[0]))
                else :
                    ner_model_2_viz = st.selectbox("Select a NER model.",index=model_selection.index('en.ner'))

        active_visualizers = visualizers
        if show_viz_selection: active_visualizers = st.sidebar.multiselect("Visualizers",options=visualizers,default=visualizers,key=key)

        all_models = ner_model_2_viz + ' en.dep.typed '  if 'dependency_tree' in active_visualizers  else ner_model_2_viz
        ner_pipe, tree_pipe =  None,None
        if 'ner' in active_visualizers :
            ner_pipe = pipe if pipe.nlu_ref == ner_model_2_viz else StreamlitUtilsOS.get_pipe(ner_model_2_viz)
            NERStreamlitBlock.visualize_ner(ner_pipe, text, generate_code_sample=show_code_snippets, key=key, show_model_select=False, show_text_input=True, show_logo=False, show_infos=False)
        if 'dependency_tree' in active_visualizers :
            tree_pipe = StreamlitUtilsOS.get_pipe('en.dep.typed') # if not ValidateVizPipe.viz_tree_satisfied(pipe) else pipe
            DepTreeStreamlitBlock.visualize_dep_tree(tree_pipe, text, generate_code_sample=show_code_snippets, key=key, show_infos=False, show_logo=False)
        if 'token_features' in active_visualizers:
            ner_pipe = pipe if pipe.nlu_ref == ner_model_2_viz else StreamlitUtilsOS.get_pipe(ner_model_2_viz)
            TokenFeaturesStreamlitBlock.visualize_tokens_information(ner_pipe, text, generate_code_sample=show_code_snippets, key=key, model_select_position=model_select_position, show_infos=False, show_logo=False, )
        if 'classification' in active_visualizers:
            ner_pipe = pipe if pipe.nlu_ref == ner_model_2_viz else StreamlitUtilsOS.get_pipe(ner_model_2_viz)
            ClassifierStreamlitBlock.visualize_classes(ner_pipe, text, generate_code_sample=show_code_snippets, key=key, model_select_position=model_select_position, show_infos=False, show_logo=False)
        if 'similarity' in active_visualizers:
            ner_pipe = pipe if pipe.nlu_ref == ner_model_2_viz else StreamlitUtilsOS.get_pipe(ner_model_2_viz)
            WordSimilarityStreamlitBlock.display_word_similarity(ner_pipe, similarity_texts,generate_code_sample=show_code_snippets, model_select_position=model_select_position, show_infos=False,show_logo=False, num_cols=num_similarity_cols)
        if 'manifold' in active_visualizers :
            ner_pipe = pipe if ner_model_2_viz in pipe.nlu_ref.split(' ')  else StreamlitUtilsOS.get_pipe(ner_model_2_viz)
            WordEmbeddingManifoldStreamlitBlock.viz_streamlit_word_embed_manifold(ner_pipe, similarity_texts, generate_code_sample=show_code_snippets, model_select_position=model_select_position, show_infos=False, show_logo=False, num_cols=num_manifold_cols)

        models_to_display_info_for = []
        if ner_pipe  is not None : models_to_display_info_for .append(ner_pipe)
        if tree_pipe is not None : models_to_display_info_for .append(tree_pipe)
        if show_models_info      :StreamlitVizTracker.display_model_info(all_models, models_to_display_info_for)
        if display_infos         : StreamlitVizTracker.display_footer()

    @staticmethod
    def viz_streamlit_word_embed_manifold(
            pipe, # nlu pipe
            default_texts: List[str] = ("Donald Trump likes to party!", "Angela Merkel likes to party!", 'Peter HATES TO PARTTY!!!! :('),
            title: Optional[str] = "Lower dimensional Manifold visualization for word embeddings",
            sub_title: Optional[str] = "Apply any of the 11 `Manifold` or `Matrix Decomposition` algorithms to reduce the dimensionality of `Word Embeddings` to `1-D`, `2-D` and `3-D` ",
            write_raw_pandas : bool = False ,
            default_algos_to_apply : List[str] = ("TSNE", "PCA"),#,'LLE','Spectral Embedding','MDS','ISOMAP','SVD aka LSA','DictionaryLearning','FactorAnalysis','FastICA','KernelPCA',),  # LatentDirichletAllocation 'NMF',
            target_dimensions : List[int] = (1,2,3),
            show_algo_select : bool = True,
            show_embed_select : bool = True,
            show_color_select: bool = True,
            MAX_DISPLAY_NUM:int=100,
            display_embed_information:bool=True,
            set_wide_layout_CSS:bool=True,
            num_cols: int = 3,
            model_select_position:str = 'side', # side or main
            key:str = "NLU_streamlit",
            additional_classifiers_for_coloring:List[str]=['pos', 'sentiment'],
            generate_code_sample:bool = False,
            show_infos:bool = True,
            show_logo:bool = True,
            n_jobs: Optional[int] = 3, # False
    ): WordEmbeddingManifoldStreamlitBlock.viz_streamlit_word_embed_manifold(
        pipe,
        default_texts,
        title,
        sub_title,
        write_raw_pandas,
        default_algos_to_apply,
        target_dimensions,
        show_algo_select,
        show_embed_select,
        show_color_select,
        MAX_DISPLAY_NUM,
        display_embed_information,
        set_wide_layout_CSS,
        num_cols,
        model_select_position,
        key,
        additional_classifiers_for_coloring,
        generate_code_sample,
        show_infos,
        show_logo,
        n_jobs,
    )


    @staticmethod
    def visualize_dep_tree(
            pipe, #nlu pipe
            text:str = 'Billy likes to swim',
            title: Optional[str] = "Dependency Parse & Part-of-speech tags",
            sub_title: Optional[str] = 'POS tags define a `grammatical label` for `each token` and the `Dependency Tree` classifies `Relations between the tokens` ',
            set_wide_layout_CSS:bool=True,
            generate_code_sample:bool = False,
            key = "NLU_streamlit",
            show_infos:bool = True,
            show_logo:bool = True,
            show_text_input:bool = True,
    ): DepTreeStreamlitBlock.visualize_dep_tree(pipe,
                                                text,
                                                title,
                                                sub_title,
                                                set_wide_layout_CSS,
                                                generate_code_sample,
                                                key,show_infos,
                                                show_logo,
                                                show_text_input,)

    @staticmethod
    def display_word_similarity(
            pipe, #nlu pipe
            default_texts: Tuple[str, str] = ("Donald Trump likes to party!", "Angela Merkel likes to party!"),
            threshold: float = 0.5,
            title: Optional[str] = "Embeddings Similarity Matrix &  Visualizations  ",
            sub_tile :Optional[str]="Visualize `word-wise similarity matrix` and calculate `similarity scores` for `2 texts` and every `word embedding` loaded",
            write_raw_pandas : bool = False,
            display_embed_information:bool = True,
            similarity_matrix = True,
            show_algo_select : bool = True,
            dist_metrics:List[str]  =('cosine'),
            set_wide_layout_CSS:bool=True,
            generate_code_sample:bool = False,
            key:str = "NLU_streamlit",
            num_cols:int=2,
            display_scalar_similarities : bool = False ,
            display_similarity_summary:bool = False,
            model_select_position:str = 'side' , # main or side
            show_infos:bool = True,
            show_logo:bool = True,
    ):WordSimilarityStreamlitBlock.display_word_similarity(pipe,
                                                           default_texts,
                                                           threshold,
                                                           title,
                                                           sub_tile,
                                                           write_raw_pandas,
                                                           display_embed_information,
                                                           similarity_matrix,
                                                           show_algo_select,
                                                           dist_metrics,
                                                           set_wide_layout_CSS,
                                                           generate_code_sample,
                                                           key,
                                                           num_cols,
                                                           display_scalar_similarities,
                                                           display_similarity_summary,
                                                           model_select_position,
                                                           show_infos,
                                                           show_logo,)








    @staticmethod
    def visualize_tokens_information(
            pipe, # nlu pipe
            text:str,
            title: Optional[str] = "Token Features",
            sub_title: Optional[str] ='Pick from `over 1000+ models` on the left and `view the generated features`',
            show_feature_select:bool =True,
            features:Optional[List[str]] = None,
            full_metadata: bool = True,
            output_level:str = 'token',
            positions:bool = False,
            set_wide_layout_CSS:bool=True,
            generate_code_sample:bool = False,
            key = "NLU_streamlit",
            show_model_select = True,
            model_select_position:str = 'side' , # main or side
            show_infos:bool = True,
            show_logo:bool = True,
            show_text_input:bool = True,
    ) -> None: TokenFeaturesStreamlitBlock.visualize_tokens_information(
        pipe,
        text,
        title,
        sub_title,
        show_feature_select,
        features,
        full_metadata,
        output_level,
        positions,
        set_wide_layout_CSS,
        generate_code_sample,
        key,
        show_model_select,
        model_select_position,
        show_infos,
        show_logo,
        show_text_input,
    )


    @staticmethod
    def visualize_ner(
            pipe, # Nlu pipe
            text:str,
            ner_tags: Optional[List[str]] = None,
            show_label_select: bool = True,
            show_table: bool = False,
            title: Optional[str] = "Named Entities",
            sub_title: Optional[str] = "Recognize various `Named Entities (NER)` in text entered and filter them. You can select from over `100 languages` in the dropdown.",
            colors: Dict[str, str] = {},
            show_color_selector: bool = False,
            set_wide_layout_CSS:bool=True,
            generate_code_sample:bool = False,
            key = "NLU_streamlit",
            model_select_position:str = 'side',
            show_model_select : bool = True,
            show_text_input:bool = True,
            show_infos:bool = True,
            show_logo:bool = True,

    ): NERStreamlitBlock.visualize_ner(
        pipe,
        text,
        ner_tags,
        show_label_select,
        show_table,
        title,
        sub_title,
        colors,
        show_color_selector,
        set_wide_layout_CSS,
        generate_code_sample,
        key,
        model_select_position,
        show_model_select,
        show_text_input,
        show_infos,
        show_logo,
    )


    @staticmethod
    def visualize_classes(
            pipe, # nlu pipe
            text:Union[str,list,pd.DataFrame, pd.Series, List[str]]=('I love NLU and Streamlit and sunny days!', 'I hate rainy daiys','CALL NOW AND WIN 1000$M'),
            output_level:Optional[str]='document',
            title: Optional[str] = "Text Classification",
            sub_title: Optional[str] = 'View predicted `classes` and `confidences` for `hundreds of text classifiers` in `over 200 languages`',
            metadata : bool = False,
            positions : bool = False,
            set_wide_layout_CSS:bool=True,
            generate_code_sample:bool = False,
            key:str = "NLU_streamlit",
            show_model_selector : bool = True ,
            model_select_position:str = 'side' ,
            show_infos:bool = True,
            show_logo:bool = True,
    )->None: ClassifierStreamlitBlock.visualize_classes(
        pipe,
        text,
        output_level,
        title,
        sub_title,
        metadata,
        positions,
        set_wide_layout_CSS,
        generate_code_sample,
        key,
        show_model_selector,
        model_select_position,
        show_infos,
        show_logo,
    )



